//! A bin that calculates the derivation of an input value.
//!
//! ```text
//!   ┌────[derivation]────┐
//!  ⇒│input         output│⇒
//!   └────────────────────┘
//! ```
//!
//! The derivation is the difference from the previous value to the
//! current value. If either the previous or the current value is
//! `Nothing`, then the output will also be `Nothing`.

use super::{
    sink_names_input, source_names_output, BinBuildEnvironment,
    BinDescription, Calculator, FetchItem, GetCalibration, Item,
    Iteration, Result, Scope, SinkBin, SinkNames, SourceBin, SourceId,
    SourceNames, SourceSinkBinDescription, WriteDotSimple, SINK_INPUT,
    SOURCE_OUTPUT,
};
use crate::error;
use indexmap::IndexSet;

static BIN_TYPE: &str = "derivation";

/// A bin that calculates the derivation of an input value.
#[derive(Debug)]
pub struct Bin {
    scope: Scope,
    source_input: Box<dyn FetchItem>,
    previous_value: Item,
    result_output: Item,
}

impl SinkBin for Bin {}

impl SourceBin for Bin {
    fn get_source_data(&self, source: &SourceId) -> Result<Item> {
        if source.id == SOURCE_OUTPUT {
            Ok(self.result_output.clone())
        } else {
            error::MissingSourceName {
                scope: self.scope.clone(),
                name: source.id.to_string(),
                bin_type: BIN_TYPE.to_string(),
            }
            .fail()
        }
    }
}

impl Calculator for Bin {
    fn calculate(&mut self, _iteration: &Iteration) -> Result<()> {
        let input = self.source_input.fetch_item(&self.scope)?;

        self.result_output =
            match (self.previous_value.to_float(), input.to_float()) {
                (Ok(p), Ok(v)) => Item::F64(v - p),
                _ => Item::Nothing,
            };
        self.previous_value = input;
        Ok(())
    }
}

/// Description for the derivation bin.
#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Description;

impl BinDescription for Description {
    type Bin = Bin;

    fn check_validity(
        &self,
        _scope: &Scope,
        _get_calibration: &mut dyn GetCalibration,
    ) -> Result<()> {
        Ok(())
    }

    fn bin_type(&self) -> &'static str {
        BIN_TYPE
    }
}

impl SinkNames for Description {
    fn sink_names(&self) -> IndexSet<String> {
        sink_names_input()
    }
}

impl SourceNames for Description {
    fn source_names(&self) -> Result<IndexSet<String>> {
        Ok(source_names_output())
    }
}

impl SourceSinkBinDescription for Description {
    fn build_bin(
        &self,
        scope: &Scope,
        env: &mut dyn BinBuildEnvironment,
    ) -> Result<Self::Bin> {
        Ok(Bin {
            scope: scope.clone(),
            source_input: env.resolve(SINK_INPUT)?,
            previous_value: Item::Nothing,
            result_output: Item::Nothing,
        })
    }
}

impl WriteDotSimple for Description {}

#[cfg(test)]
mod tests {
    use super::Description;
    use crate::bins::{directsource, verificationsink};
    use crate::{run_bin, Item, Result};
    use indexmap::indexset;

    #[test]
    fn simulate() -> Result<()> {
        use crate::Item::*;

        let input = directsource::Description {
            columns: indexset!["input".to_string()],
            rows: vec![
                vec![Nothing],
                vec![Item::from(0f64)],
                vec![Item::from(2f64)],
                vec![Item::from(1.5f64)],
                vec![Item::from(0f64)],
                vec![Nothing],
                vec![Item::from(-3i32)],
                vec![Item::from(-3i32)],
                vec![Item::from(-4.5f32)],
                vec![Nothing],
                vec![Item::from(-3)],
                vec![Item::from(5)],
            ]
            .into(),
        };
        let verification = verificationsink::Description {
            columns: indexset!["output".to_string()],
            expected: vec![
                vec![Nothing],
                vec![Nothing],
                vec![Item::from(2f64)],
                vec![Item::from(-0.5f64)],
                vec![Item::from(-1.5f64)],
                vec![Nothing],
                vec![Nothing],
                vec![Item::from(0f64)],
                vec![Item::from(-1.5f64)],
                vec![Nothing],
                vec![Nothing],
                vec![Item::from(8f64)],
            ]
            .into(),
        };

        run_bin(&input, &Description {}, &verification)
    }
}
