//! A bin that multiplies two values.
//!
//! ```text
//!   ┌────[multiply]────┐
//!  ⇒│a           output│⇒
//!  ⇒│b                 │
//!   └──────────────────┘
//! ```

use super::{
    sink_names_a_b, source_names_output, BinBuildEnvironment,
    BinDescription, Calculator, FetchItem, GetCalibration, Item,
    Iteration, Result, Scope, SinkBin, SinkNames, SourceBin, SourceId,
    SourceNames, SourceSinkBinDescription, WriteDotSimple, SINK_A, SINK_B,
    SOURCE_OUTPUT,
};
use crate::error;
use indexmap::IndexSet;

static BIN_TYPE: &str = "multiply";

/// A bin that multiplies two values.
#[derive(Debug)]
pub struct Bin {
    scope: Scope,

    source_a: Box<dyn FetchItem>,
    source_b: Box<dyn FetchItem>,

    result_output: Item,
}

impl SinkBin for Bin {}

impl SourceBin for Bin {
    fn get_source_data(&self, source: &SourceId) -> Result<Item> {
        if source.id == SOURCE_OUTPUT {
            Ok(self.result_output.clone())
        } else {
            error::MissingSourceName {
                scope: self.scope.clone(),
                name: source.id.to_string(),
                bin_type: BIN_TYPE.to_string(),
            }
            .fail()
        }
    }
}

impl Calculator for Bin {
    fn calculate(&mut self, _iteration: &Iteration) -> Result<()> {
        let a = self.source_a.fetch_item(&self.scope)?;
        let b = self.source_b.fetch_item(&self.scope)?;

        self.result_output = match (a.to_float(), b.to_float()) {
            (Ok(a), Ok(b)) => Item::F64(a * b),
            _ => Item::Nothing,
        };
        Ok(())
    }
}

/// Description for the multiply bin.
#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Description;

impl BinDescription for Description {
    type Bin = Bin;

    fn check_validity(
        &self,
        _scope: &Scope,
        _get_calibration: &mut dyn GetCalibration,
    ) -> Result<()> {
        Ok(())
    }

    fn bin_type(&self) -> &'static str {
        BIN_TYPE
    }
}

impl SinkNames for Description {
    fn sink_names(&self) -> IndexSet<String> {
        sink_names_a_b()
    }
}

impl SourceNames for Description {
    fn source_names(&self) -> Result<IndexSet<String>> {
        Ok(source_names_output())
    }
}

impl SourceSinkBinDescription for Description {
    fn build_bin(
        &self,
        scope: &Scope,
        env: &mut dyn BinBuildEnvironment,
    ) -> Result<Self::Bin> {
        Ok(Bin {
            scope: scope.clone(),
            source_a: env.resolve(SINK_A)?,
            source_b: env.resolve(SINK_B)?,
            result_output: Item::Nothing,
        })
    }
}

impl WriteDotSimple for Description {}

#[cfg(test)]
mod tests {
    use super::Description;
    use crate::bins::{directsource, verificationsink};
    use crate::Item as I;
    use crate::{run_bin, Result};
    use indexmap::indexset;

    #[test]
    fn simulate() -> Result<()> {
        use crate::Item::*;

        let input = directsource::Description {
            columns: indexset!["a".to_string(), "b".to_string()],
            rows: vec![
                vec![Nothing, Nothing],
                vec![Nothing, I::from(1f64)],
                vec![I::from(1f64), Nothing],
                vec![I::from(1f64), I::from(1f64)],
                vec![I::from(0f64), I::from(1f64)],
                vec![I::from(1f64), I::from(0f64)],
                vec![I::from(0f64), I::from(0f64)],
                vec![I::from(2f64), I::from(0f64)],
                vec![I::from(1f64), I::from(2f64)],
                vec![I::from(2f64), I::from(1f64)],
                vec![I::from(2f64), I::from(2f64)],
            ]
            .into(),
        };
        let verification = verificationsink::Description {
            columns: indexset!["output".to_string()],
            expected: vec![
                vec![Nothing],
                vec![Nothing],
                vec![Nothing],
                vec![I::from(1f64)],
                vec![I::from(0f64)],
                vec![I::from(0f64)],
                vec![I::from(0f64)],
                vec![I::from(0f64)],
                vec![I::from(2f64)],
                vec![I::from(2f64)],
                vec![I::from(4f64)],
            ]
            .into(),
        };

        run_bin(&input, &Description {}, &verification)
    }
}
